using System;
using UnityEngine;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Vinput
{
    public class AutoThruster : MVRScript
    {
        public bool __DEBUG__ = false;

        // ============================================================================================ //
        // ======================================== PROPERTIES ======================================== //
        // ============================================================================================ //
        // UI variables
        private JSONStorableBool _activeJSON;
        private JSONStorableStringChooser _controlJSON;
        private JSONStorableBool _thrustEnabledJSON;
        private JSONStorableFloat _distanceJSON;
        private JSONStorableFloat _distance0JSON;
        private JSONStorableFloat _distanceRandomJSON;
        private JSONStorableStringChooser _axisJSON;
        private JSONStorableFloat _thrustTimeOffsetJSON;
        private JSONStorableBool _rotEnabledJSON;
        private JSONStorableFloat _rotTimeOffsetJSON;
        private JSONStorableFloat _rotStartXJSON;
        private JSONStorableFloat _rotStartYJSON;
        private JSONStorableFloat _rotStartZJSON;
        private JSONStorableFloat _rotEndXJSON;
        private JSONStorableFloat _rotEndYJSON;
        private JSONStorableFloat _rotEndZJSON;
        private JSONStorableBool _rotLinearJSON;
        private JSONStorableFloat _dirOffsetXJSON;
        private JSONStorableFloat _dirOffsetYJSON;
        private JSONStorableFloat _dirOffsetZJSON;

        private JSONStorableBool _useMasterJSON;

        // Person atoms, controls, rigidbodies
        private FreeControllerV3 _control;

        // Internal
        private Cycler _cycler;
        private SyncManager _sync;
        private JSONStorableFloat _cyclePctInJSON;
        private JSONStorableStringChooser _easingInJSON;
        private JSONStorableStringChooser _easingOutJSON;
        private static readonly Dictionary<string, Vector3> _axisVectors = new Dictionary<string, Vector3>
        {
            {"X", Vector3.right},
            {"Y", Vector3.up},
            {"Z", Vector3.forward}
        };
        private float _thrustDistance => _distanceJSON.val - _distance0JSON.val;
        private CtrlSnapshot _ctrlSnapshot = null;
        private CtrlSnapshot _parentSnapshot = null;
        private bool _restoring = false;
        private float _distanceFactor = 1.0f;

        // Sync variables
        public JSONStorableBool isActive = new JSONStorableBool("isActive", false);
        public JSONStorableFloat cyclePct = new JSONStorableFloat("cyclePct", 0f, 0f, 1f);
        private float _lastPct = 0f;



        // ============================================================================================ //
        // ======================================== METHODS =========================================== //
        // ============================================================================================ //

        // BUILT-IN
        // Called by game engine on initialization
        public override void Init()
        {
            try
            {
                _cycler = new Cycler(this, OnThrustBegin, OnThrustUpdate);
                _sync = new SyncManager(this);
                RegisterBool(isActive);
                RegisterFloat(cyclePct);
                SetupUI();
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(AutoThruster)}.{nameof(Init)}: {e}");
            }
        }


        // Called by game engine before first update
        private void Start() { _UpdateState(); }

        // Events/callbacks
        public void OnEnable() { _UpdateState(); }
        public void OnDisable() { _Deactivate(); }
        public void OnDestroy()
        {
            if (_cycler != null && _cycler.IsActive) _cycler.StopCycle();
            _RestoreCtrlSnapshot(_ctrlSnapshot);
        }

        // PLUGIN
        // Cycler functions

        private void OnThrustBegin()
        {
            if (_useMasterJSON.val) return;
            _distanceFactor = UnityEngine.Random.Range(1 - _distanceRandomJSON.val, 1 + _distanceRandomJSON.val);
        }
        private void OnThrustUpdate(float pct)
        {
            cyclePct.val = pct;
            if (_useMasterJSON.val)
            {
                if (_sync.masterActive)
                {
                    pct = _sync.CurrentPct;
                    if (pct < _lastPct) _distanceFactor = UnityEngine.Random.Range(1 - _distanceRandomJSON.val, 1 + _distanceRandomJSON.val);
                    _lastPct = pct;
                }
                else return;
            }
            if (__DEBUG__) SuperController.LogMessage($"Updating thrust: pct = {pct}, t = {_T((pct + _thrustTimeOffsetJSON.val + 1) % 1, _easingInJSON.val, _easingOutJSON.val)}");

            Vector3 originPos = _ctrlSnapshot.localPosition;
            Quaternion originRot = _ctrlSnapshot.localRotation;

            Vector3 direction = (originRot * Quaternion.Euler(_dirOffsetXJSON.val, _dirOffsetYJSON.val, _dirOffsetZJSON.val) * _axisVectors[_axisJSON.val]).normalized;
            Vector3 startPos = originPos + direction * _distance0JSON.val / 10;
            Vector3 endPos = startPos + direction * _thrustDistance * _distanceFactor / 10;

            if (_thrustEnabledJSON.val)
                _control.transform.localPosition = Vector3.LerpUnclamped(startPos, endPos, _T((pct + _thrustTimeOffsetJSON.val + 1) % 1, _easingInJSON.val, _easingOutJSON.val));

            if (_rotEnabledJSON.val)
            {
                float _t = _rotLinearJSON.val ? _T((pct + _rotTimeOffsetJSON.val + 1) % 1, "Linear", "Linear") : _T((pct + _rotTimeOffsetJSON.val + 1) % 1, _easingInJSON.val, _easingOutJSON.val);
                _control.transform.localRotation = originRot * Quaternion.LerpUnclamped(
                   Quaternion.Euler(_rotStartXJSON.val, _rotStartYJSON.val, _rotStartZJSON.val),
                   Quaternion.Euler(_rotEndXJSON.val, _rotEndYJSON.val, _rotEndZJSON.val),
                   _t
               );
            }

        }

        private float _T(float currPct, string easeIn = null, string easeOut = null)
        {
            easeIn = easeIn ?? "Linear";
            easeOut = easeOut ?? "Linear";
            bool r = currPct >= _cyclePctInJSON.val;
            float x = r ? (currPct - _cyclePctInJSON.val) / (1 - _cyclePctInJSON.val) : currPct / _cyclePctInJSON.val;
            return r ? 1 - Easings.EasingFuncs[easeOut](x) : Easings.EasingFuncs[easeIn](x);
        }


        // Activation/deactivation
        private void _UpdateState()
        {
            if (!_restoring)
            {
                if (_activeJSON.val && enabled && (_thrustEnabledJSON.val || _rotEnabledJSON.val))
                {
                    // Try to activate
                    _control = containingAtom.freeControllers.First(fc => fc.name == _controlJSON.val);
                    if (_ctrlSnapshot == null || _ctrlSnapshot.name != _control.name)
                    {
                        _RestoreCtrlSnapshot(_ctrlSnapshot);
                        _ctrlSnapshot = new CtrlSnapshot(_control);
                    }

                    _cycler.StartCycle();
                    isActive.val = true;

                }
                else _Deactivate();
            }
        }

        private void _Deactivate()
        {
            _distanceFactor = 1.0f;
            isActive.SetVal(false);
            if (!_cycler.IsActive) return;

            _cycler.StopCycle();
            _RestoreCtrlSnapshot(_control);
            _ctrlSnapshot = null;
        }
        private void OnControlChange(string _) { _UpdateState(); }
        private void OnActiveChange(bool _) { _UpdateState(); }
        private void OnEnableRotationsChange(bool val)
        {
            if (!val && _cycler.IsActive) _control.transform.localRotation = _ctrlSnapshot.localRotation;
        }
        private void OnMasterAtomChange(string uid)
        {
            _sync._masterPluginJSON.SetValToDefault();
            _UpdateState();
        }
        private void OnMasterPluginChange(string pid) { _UpdateState(); }

        // Save/restore
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var json = base.GetJSON(includePhysical, includeAppearance, forceStore);

            if (_ctrlSnapshot != null) json["CtrlSnapshot"] = Serializer.Serialize(_ctrlSnapshot);
            if (_parentSnapshot != null) json["ParentSnapshot"] = Serializer.Serialize(_parentSnapshot);

            needsStore = true;
            return json;
        }
        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            _restoring = true;
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            if (jc.HasKey("CtrlSnapshot")) _ctrlSnapshot = Serializer.DeserializeCtrlSnapshot(jc["CtrlSnapshot"] as JSONClass);
            if (jc.HasKey("ParentSnapshot")) _parentSnapshot = Serializer.DeserializeCtrlSnapshot(jc["ParentSnapshot"] as JSONClass);

            _restoring = false;
        }
        private void _RestoreCtrlSnapshot(FreeControllerV3 control)
        {
            if (control == null) return;

            if (control.name == _ctrlSnapshot.name) _RestoreCtrlSnapshot(_ctrlSnapshot);
        }

        private void _RestoreCtrlSnapshot(CtrlSnapshot snapshot)
        {
            if (snapshot == null) return;

            FreeControllerV3 ctrl = containingAtom.freeControllers.First(fc => fc.name == snapshot.name);
            if (ctrl != null)
            {
                ctrl.transform.localPosition = snapshot.localPosition;
                ctrl.transform.localRotation = snapshot.localRotation;
            }
        }

        // UI

        private void SetupUI()
        {
            string instructionText =
                "<color=#202020><size=36><b>Settings Information</b></size></color>\n\n" +
                "<b>Control</b>: Name of the control to thrust.\n" +
                "<b>Thrust Axis</b>: The axis of the control along which to thrust (X = Red, Y = Green, Z = Blue).\n\n" +

                "<color=#101060><b>Cycle Settings</b></color>\n" +
                "<b>Cycle Duration:</b> Duration of a complete <i>In</i> and <i>Out</i> motion.\n" +
                "<b>Cycle Duration % In:</b> Proportion of Cycle Duration taken by the <i>In</i> motion.\n" +
                "0.5 = <i>In</i> and <i>Out</i> have equal duration.\n" +
                "Less than 0.5 = <i>In</i> motion is faster than <i>Out</i> motion.\n" +
                "More than 0.5 = <i>In</i> motion is slower than <i>Out</i> motion.\n" +
                "<b>Cycle Duration Random %:</b> Proportion of Cycle Duration to randomly add/subtract from each cycle. A value of 0.0 means no random variation, a value of 0.5 means cycles will randomly vary between 50% and 150% of Cycle Duration.\n" +
                "<b>Ease In:</b> Easing function for <i>In</i> motion. See https://easings.net for reference.\n" +
                "<b>Ease Out:</b> Easing function for <i>Out</i> motion.\n\n" +

                "<color=#101060><b>Thrust Settings</b></color>\n" +
                "<b>Distance</b>: The distance to thrust. Set this to a negative value to thrust in the negative direction of the Thrust Axis.\n" +
                "<b>Distance Random %</b>: Proportion of Distance to randomly add/subtract from each cycle. A value of 0.0 means no random variation, a value of 0.5 means distances will randomly vary between 50% and 150% of Distance.\n" +
                "<b>Thrust Time Offset %</b>: Positive values 'shift' the thrust motions forward in time, negative values 'shift' them backwards.\n" +
                "<b>Direction X/Y/Z Offset</b>: Adjustments to the thrust direction.\n\n" +

                "<color=#101060><b>Rotation Settings</b></color>\n" +
                "<b>Rotation X/Y/Z Start</b>: Relative rotations at the start of the thrust.\n" +
                "<b>Rotation X/Y/Z End</b>: Relative rotations at the end of the thrust.\n" +
                "<b>Rotation Time Offset %</b>: Positive values 'shift' the rotations forward in time, negative values 'shift' them backwards.\n" +
                "<b>Linear rotation</b>: Use linear easing for rotations instead of the one specified in Ease In/Out.\n\n" +

                "<color=#101060><b>Sync Settings</b></color>\n" +
                "When active, use the cycle timing from the specified atom and plugin instance. Cycle Duration and Cycle Duration Random % from this instance will be ignored.\n" +
                "All other settings still apply, such as Cycle Duration % In, easings, and distance settings.\n\n";

            _activeJSON = UI_Utils.SetupToggle(this, "<color=#1010A0><size=32><b>Active</b></size></color>", false);
            _activeJSON.setCallbackFunction += OnActiveChange;


            // Control
            _controlJSON = new JSONStorableStringChooser("ControlName", Scene_Utils.ControlNames(containingAtom), Scene_Utils.ControlNames(containingAtom)[0], "Control");
            RegisterStringChooser(_controlJSON);
            _controlJSON.setCallbackFunction += OnControlChange;
            CreateFilterablePopup(_controlJSON).popup.onOpenPopupHandlers += () => { _controlJSON.choices = Scene_Utils.ControlNames(containingAtom); };
            CreateSpacer(true).height = 50f;
            _axisJSON = new JSONStorableStringChooser("ThrustAxis", _axisVectors.Keys.ToList(), "Z", "Thrust Axis");
            RegisterStringChooser(_axisJSON);
            CreateScrollablePopup(_axisJSON, true).popup.onOpenPopupHandlers += () => { _axisJSON.choices = _axisVectors.Keys.ToList(); };

            // Cycler
            CreateSpacer().height = 5f;
            _cycler.SetupUI(this);
            _cyclePctInJSON = UI_Utils.SetupSlider(this, "Duration % In", 0.5f, 0.1f, 0.9f);
            _easingInJSON = new JSONStorableStringChooser("EaseIn", Easings.EasingTypes, "Linear", "Ease In");
            CreateScrollablePopup(_easingInJSON);
            RegisterStringChooser(_easingInJSON);
            _easingOutJSON = new JSONStorableStringChooser("EaseOut", Easings.EasingTypes, "Linear", "Ease Out");
            CreateScrollablePopup(_easingOutJSON);
            RegisterStringChooser(_easingOutJSON);

            // Instructions
            CreateSpacer(true).height = 25f;
            CreateTextField(new JSONStorableString("text", instructionText), true).height = 620;

            // Thrust
            CreateSpacer().height = 10f;
            _thrustEnabledJSON = UI_Utils.SetupToggle(this, "Enable Thrust", true);
            _thrustEnabledJSON.setCallbackFunction += OnActiveChange;
            _distance0JSON = UI_Utils.SetupSlider(this, "Distance Start", 0.0f, -5.0f, 5.0f, false);
            _distanceJSON = UI_Utils.SetupSlider(this, "Distance End", 2.0f, -5.0f, 5.0f, false);
            _distanceRandomJSON = UI_Utils.SetupSlider(this, "Distance Random %", 0f, 0f, 0.5f, false);
            _thrustTimeOffsetJSON = UI_Utils.SetupSlider(this, "Thrust Time Offset %", 0.0f, -0.5f, 0.5f);
            CreateSpacer(true).height = 75f;
            _dirOffsetXJSON = UI_Utils.SetupSlider(this, "Direction Offset X", 0f, -90f, 90f, false, true);
            _dirOffsetYJSON = UI_Utils.SetupSlider(this, "Direction Offset Y", 0f, -90f, 90f, false, true);
            _dirOffsetZJSON = UI_Utils.SetupSlider(this, "Direction Offset Z", 0f, -90f, 90f, false, true);


            // Settings - Rotations
            CreateSpacer().height = 10f;
            CreateSpacer(true).height = 210f;
            _rotEnabledJSON = UI_Utils.SetupToggle(this, "Enable Rotations", true);
            _rotEnabledJSON.setCallbackFunction += OnEnableRotationsChange;
            _rotStartXJSON = UI_Utils.SetupSlider(this, "Rotation X Start", 0f, -90f, 90f, false);
            _rotEndXJSON = UI_Utils.SetupSlider(this, "Rotation X End", 0f, -90f, 90f, false, true);
            _rotStartYJSON = UI_Utils.SetupSlider(this, "Rotation Y Start", 0f, -90f, 90f, false);
            _rotEndYJSON = UI_Utils.SetupSlider(this, "Rotation Y End", 0f, -90f, 90f, false, true);
            _rotStartZJSON = UI_Utils.SetupSlider(this, "Rotation Z Start", 0f, -90f, 90f, false);
            _rotEndZJSON = UI_Utils.SetupSlider(this, "Rotation Z End", 0f, -90f, 90f, false, true);
            _rotTimeOffsetJSON = UI_Utils.SetupSlider(this, "Rotation Time Offset %", 0f, -0.5f, 0.5f);
            _rotLinearJSON = UI_Utils.SetupToggle(this, "Linear Rotation", false);

            CreateSpacer().height = 10f;
            _useMasterJSON = UI_Utils.SetupToggle(this, "Sync with Master", false);
            _useMasterJSON.setCallbackFunction += OnActiveChange;
            _sync.SetupUI(this);
            CreateSpacer().height = 400f;
        }
    }
}
